
from zope import schema
from zope.interface import Interface


class IElasticSettings(Interface):
    only_published = schema.Bool(
        title=u'Index only published content?',
        default=True)
    index_name = schema.TextLine(
        title=u"Index name",
        required=True)
    server_urls = schema.List(
        title=u"Server URLs",
        value_type=schema.URI(),
        required=True)
    public_server_urls = schema.List(
        title=u"Server URLs to use for the public search",
        description=u"If not specified regular server URLs will be used",
        value_type=schema.URI(),
        required=True)






from plone.app.controlpanel.form import ControlPanelForm
from zope.formlib import form
from zope.i18nmessageid import MessageFactory

from collective.elasticindex.interfaces import IElasticSettings
from collective.elasticindex.changes import changes
from collective.elasticindex.utils import create_index


_ = MessageFactory('collective.elasticindex')


class ElasticIndexSettings(ControlPanelForm):
    label = _("Elasticsearch configuration")
    description = _("Configure elasticsearch index.")
    form_name = _("Elasticsearch configuration")
    form_fields = form.Fields(IElasticSettings)

    @form.action('Create index', name='create_index')
    def create_index(self, action, data):
        create_index(IElasticSettings(self.context))


    @form.action('Import site content', name='import_content')
    def import_site_content(self, action, data):
        changes.verify_and_index_container(self.context)

    actions += form.Actions(*list(ControlPanelForm.actions))





from Acquisition import aq_base
from Products.CMFCore.interfaces import IPropertiesTool
from zope.component import getUtility


class EmptySettings(object):
    only_published = True
    index_name = None
    server_urls = []
    public_server_urls = []


class SettingsAdapter(object):

    def __init__(self, context):
        self._id = context.getId()
        properties = getUtility(IPropertiesTool)
        self._activated = hasattr(
            aq_base(properties),
            'elasticindex_properties')
        if self._activated:
            self._properties = properties.elasticindex_properties
        else:
            self._properties = EmptySettings()

    @property
    def activated(self):
        return self._activated

    @apply
    def only_published():

        def getter(self):
            return bool(self._properties.only_published)

        def setter(self, value):
            self._properties.only_published = bool(value)
            return value

        return property(getter, setter)

    @apply
    def index_name():

        def getter(self):
            if self._properties.index_name:
                return self._properties.index_name
            return self._id

        def setter(self, value):
            self._properties.index_name = value
            return value

        return property(getter, setter)

    @apply
    def server_urls():

        def getter(self):
            return self._properties.server_urls

        def setter(self, value):
            self._properties.server_urls = tuple(value)
            return self._properties.server_urls

        return property(getter, setter)

    @apply
    def public_server_urls():

        def getter(self):
            return self._properties.public_server_urls

        def setter(self, value):
            self._properties.public_server_urls = tuple(value)
            return self._properties.public_server_urls

        return property(getter, setter)




from Products.CMFCore.interfaces import IFolderish, IContentish
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_unicode
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from transaction.interfaces import ISavepointDataManager, IDataManagerSavepoint
from plone.i18n.normalizer.base import mapUnicode
from zope.component import queryUtility
from zope.interface import implements
import threading
import transaction
import logging
import re

from Acquisition import aq_base

from collective.elasticindex.interfaces import IElasticSettings
from collective.elasticindex.utils import connect

logger = logging.getLogger('collective.elasticindex')

num_sort_regex = re.compile('\d+')

def sortable_string(string):
    return num_sort_regex.sub(
        lambda m: m.group().zfill(6),
        mapUnicode(safe_unicode(string)).lower().strip())

def get_uid(content):
    """Return content identifier to use in ES.
    """
    if IPloneSiteRoot.providedBy(content):
        uid = 'root'
    else:
        uid = content.UID()
    return uid or None

def get_data(content):
    """Return data to index in ES.
    """
    uid = get_uid(content)
    if not uid:
        return None, None
    title = content.Title()
    try:
        text = content.SearchableText()
    except:
        text = title
    data = {'title': title,
            'metaType': content.portal_type,
            'sortableTitle': sortable_string(title),
            'description': content.Description(),
            'subject': content.Subject(),
            'contributors': content.Contributors(),
            'url': content.absolute_url(),
            'author': content.Creator(),
            'content': text}

    if hasattr(aq_base(content), 'pub_date_year'):
        data['publishedYear'] = getattr(content, 'pub_date_year')

    created = content.created()
    if created is not (None, 'None'):
        data['created'] = created.strftime('%Y-%m-%dT%H:%M:%S')

    modified = content.modified()
    if modified is not (None, 'None'):
        data['modified'] = modified.strftime('%Y-%m-%dT%H:%M:%S')

    return uid, data

def list_content(content, callback):
    """Recursively list CMF content out of the given one. ``callback``
    is called every thousand items after a commit.
    """

    def recurse(content):
        for child in content.contentValues():
            if IFolderish.providedBy(child):
                for grandchild in recurse(child):
                    yield grandchild
            yield child

    count = 0
    if IFolderish.providedBy(content):
        for child in recurse(content):
            yield child
            count += 1
            if count > 1000:
                transaction.commit()
                content._p_jar.cacheGC()
                callback()
                count = 0
        yield content
    elif IContentish.providedBy(content):
        yield content


class ElasticSavepoint(object):
    implements(ISavepointDataManager)

    def __init__(self, manager, index, unindex):
        self._index = index.copy()
        self._unindex = set(unindex)

    def restore(self):
        self.manager._index = self._index
        self.manager._unindex = self._unindex


class ElasticChanges(threading.local):
    implements(IDataManagerSavepoint)

    def __init__(self, manager):
        self.manager = manager
        self._clear()

    def _clear(self):
        self._index = dict()
        self._unindex = set()
        self._settings = None
        self._connection = None
        self._activated = None
        self._get_status = None

    def _get_settings(self):
        if self._settings is None:
            portal = queryUtility(IPloneSiteRoot)
            if portal is None:
                return None
            self._settings = IElasticSettings(portal)
            if self._settings.only_published:
                self._get_status = getToolByName(
                    portal, 'portal_workflow').getInfoFor
        return self._settings

    @property
    def only_published(self):
        settings = self._get_settings()
        if settings is None:
            return False
        return self._settings.only_published

    def _is_activated(self):
        if self._activated is None:
            settings = self._get_settings()
            if settings is None:
                return False
            self._activated = bool(settings.activated)
            if self._activated:
                transaction = self.manager.get()
                transaction.join(self)
        return bool(self._activated)

    def should_index_content(self, content):
        if not self._is_activated():
            return False
        if self._get_status is None:
            return True
        return self._get_status(
            content, 'review_state', default='nope') == 'published'

    def should_index_container(self, contents):
        if self._is_activated():
            for content in contents:
                if (self._get_status is None or
                    self._get_status(
                        content, 'review_state', default='nope') == 'published'):
                    yield content

    def verify_and_index_container(self, content):
        if not self._is_activated():
            return
        for item in self.should_index_container(
            list_content(content, self._is_activated)):
            uid, data = get_data(item)
            if data:
                if uid in self._unindex:
                    del self._unindex[uid]
                self._index[uid] = data

    def index_content(self, content):
        if not self._is_activated():
            return
        uid, data = get_data(content)
        if data:
            if uid in self._unindex:
                del self._unindex[uid]
            self._index[uid] = data

    def unindex_content(self, content):
        if not self._is_activated():
            return
        uid = get_uid(content)
        if uid in self._index:
            del self._index[uid]
        self._unindex.add(uid)

    def savepoint(self):
        return ElasticSavepoint(self, self._index, self._unindex)

    def commit(self, transaction):
        pass

    def sortKey(self):
        return 'Z' * 100

    def abort(self, transaction):
        self._clear()

    def tpc_begin(self, transaction):
        pass

    def tpc_vote(self, transaction):
        if self._index or self._unindex:
            settings = self._get_settings()
            if settings.server_urls:
                self._connection = connect(settings.server_urls)

    def tpc_finish(self, transaction):
        if self._connection is not None:
            settings = self._get_settings()
            for uid, data in self._index.iteritems():
                try:
                    self._connection.index(
                        data,
                        settings.index_name,
                        'document',
                        id=uid,
                        bulk=True)
                except:
                    logger.exception(
                        'Error while indexing document in Elasticsearch')
            for uid in self._unindex:
                try:
                    self._connection.delete(
                        settings.index_name,
                        'document',
                        uid,
                        bulk=True)
                except:
                    logger.exception(
                        'Error while indexing document in Elasticsearch')
            if self._index or self._unindex:
                try:
                    self._connection.flush_bulk(True)
                except:
                    logger.exception(
                        'Error while flushing changes to Elasticsearch')
        self._clear()

    def tpc_abort(self, transaction):
        self._clear()


changes = ElasticChanges(transaction.manager)




import pyes
import urlparse

ANALYZED_STRING_MAPPING = {
    'index': 'analyzed',
    'type': 'string',
    'store': 'yes'
}

STORED_STRING_MAPPING = {
    'index': 'not_analyzed',
    'type': 'string',
    'store': 'yes',
}

STRING_MAPPING = {
    'index': 'not_analyzed',
    'type': 'string',
    'store': 'no',
}

DATE_MAPPING = {
    'index': 'not_analyzed',
    'type': 'date',
    'store': 'no'
}

INT_MAPPING = {
    'index': 'not_analyzed',
    'type': 'integer',
    'store': 'no'
}

DOCUMENT_MAPPING = {
    '_index': {
        'enabled': True
    },
    # Stored
    'title': ANALYZED_STRING_MAPPING,
    'subject': ANALYZED_STRING_MAPPING,
    'description': ANALYZED_STRING_MAPPING,
    'content': ANALYZED_STRING_MAPPING,
    'author': ANALYZED_STRING_MAPPING,
    'contributors': ANALYZED_STRING_MAPPING,

    # Not analyzed
    '_id' : STORED_STRING_MAPPING,
    'url': STORED_STRING_MAPPING,
    'metaType': STORED_STRING_MAPPING,

    # Not stored
    'created': DATE_MAPPING,
    'modified': DATE_MAPPING,
    'publishedYear': INT_MAPPING,
    'sortableTitle': STRING_MAPPING,
}

def parse_url(url):
    info = urlparse.urlparse(url)
    if ':' in info.netloc:
        url, port = info.netloc.split(':', 1)
    else:
        port = 80
        if info.scheme == 'https':
            port = 443
        url = info.netloc
    return 'http', url, int(port)


def connect(urls):
    try:
        return pyes.ES(map(parse_url, urls))
    except:
        raise ValueError('Cannot connect to servers')


def create_index(settings):
    connection = connect(settings.server_urls)
    connection.indices.create_index_if_missing(settings.index_name)
    connection.indices.put_mapping(
        'document', {'properties' : DOCUMENT_MAPPING}, [settings.index_name])




from collective.elasticindex.changes import changes
from Products.CMFCore.utils import getToolByName


def content_added(content, event):
    factorytool = getToolByName(content, 'portal_factory')
    if factorytool in content.aq_chain:
        return
    if changes.should_index_content(content):
        changes.index_content(content)


def content_modified(content, event):
    factorytool = getToolByName(content, 'portal_factory')
    if factorytool in content.aq_chain:
        return
    if changes.should_index_content(content):
        changes.index_content(content)


def content_deleted(content, event):
    if event.newParent is None:
        if changes.should_index_content(content):
            changes.unindex_content(content)


def content_published(content, event):
    if not changes.only_published:
        return
    if (event.old_state.getId() == 'published' and
        event.new_state.getId() != 'published'):
        changes.unindex_content(content)
    elif (event.new_state.getId() == 'published' and
          event.old_state.getId() != 'published'):
        changes.index_content(content)





from Products.Five.browser import BrowserView
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from collective.elasticindex.interfaces import IElasticSettings
from zope.component import getUtility
import json


class SearchPage(BrowserView):

    def update(self):
        settings = IElasticSettings(getUtility(IPloneSiteRoot))
        self.server_urls = json.dumps(list(
                settings.public_server_urls or settings.server_urls))
        self.index_name = settings.index_name
        self.expanded = ''
        if 'advanced' in self.request.form:
            self.expanded = 'expanded'
        self.request.set('disable_border', 1)
        self.request.set('disable_plone.leftcolumn', 1)
        self.request.set('disable_plone.rightcolumn', 1)

    def __call__(self):
        self.update()
        return super(SearchPage, self).__call__()



# -*- coding: utf-8 -*-
# Copyright (c) 2013 Infrae. All rights reserved.
# See also LICENSE.txt

from setuptools import setup, find_packages
import os

version = '1.0'

setup(name='collective.elasticindex',
      version=version,
      description="Index Plone content in Elastic Search",
      long_description=open("README.txt").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      # Get more strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Programming Language :: Python",
        ],
      keywords='plone index elasticsearch search',
      author='Infrae',
      author_email='info@infrae.com',
      url='',
      license='BSD',
      package_dir={'': 'src'},
      packages=find_packages('src'),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'pyes',
        'setuptools',
        'zope.component',
        'zope.formlib',
        'zope.interface',
        'zope.schema',
        ],
      )




from zope.formlib import form
from zope.interface import implements
from plone.app.portlets.portlets import base
from plone.portlets.interfaces import IPortletDataProvider
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile


class ISearchPortlet(IPortletDataProvider):
    """Search portlet.
    """


class SearchPortlet(base.Assignment):
    implements(ISearchPortlet)


class SearchPortletRenderer(base.Renderer):
    render = ViewPageTemplateFile('portlet.pt')


class SearchPortletAddForm(base.AddForm):
    """Creation form for search portlet.
    """
    form_fields = form.Fields()

    def create(self, data):
        return SearchPortlet()


class SearchPortletEditForm(base.EditForm):
    """Edit form for search portlet.
    """
    form_fields = form.Fields()
